##Calcular declives v4f=name
##Caminhos=vector line
##Multiplicador_Comprimento_segmento=number 5
##MDT=raster

from qgis.core import *
from qgis.utils import *
from PyQt4.QtGui import *
from PyQt4.QtCore import *
import processing
from osgeo import gdal


# Verificar se existe uma layer denominada SplitPoints, e caso afirmativo, remove-la
while len(QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')) > 0:
    layer = QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')[0]
    QgsMapLayerRegistry.instance().removeMapLayer(layer.id())
# Verificar se existe uma layer denominada Intersection, e caso afirmativo, remove-la
while len(QgsMapLayerRegistry.instance().mapLayersByName('Intersection')) > 0:
    layer = QgsMapLayerRegistry.instance().mapLayersByName('Intersection')[0]
    QgsMapLayerRegistry.instance().removeMapLayer(layer.id())

if (Multiplicador_Comprimento_segmento < 4):
    QMessageBox.information(None, "Nota", "O valor do Multiplicador do Comprimento do Segmento deve ser maior que 4. Aconselha-se um valor entre 4 e 10. A terminar script.")
else:
    theLayerActive = processing.getObject(Caminhos)
    iface.setActiveLayer(theLayerActive)
    # Determinar o codigo EPSG da layer escolhida e avisar caso seja WGS84
    layerCRS = theLayerActive.crs()
    lyrCRS =layerCRS.authid()
    if lyrCRS=="EPSG:4326":
        QMessageBox.information(None, "Nota", "A layer de caminhos encontra-se no sistema de coordenadas geograficas WGS84 (EPSG:4326). Projecte esta layer num sistema de coordenadas planas para poder calcular comprimentos em unidades metricas. Aconselha-se a que a layer de caminhos e a layer do MDT estejam no mesmo sistema de coordenadas planas.")
    else:
        # Despistar layers do tipo MultiPart
        iter = theLayerActive.getFeatures()
        geoms = []
        for feature in iter:
            geom = feature.geometry()
            if geom.isMultipart():
                geoms.append(geom)
        if len(geoms) > 0:
            QMessageBox.information(None, "Nota:", "Layer activa do tipo MultiPart. Converta a layer para SinglePart antes de usar este script."  )
        else:
            #  Despistar MDTs com mais de uma banda
            theRasterLayer = gdal.Open(MDT)
            bands = theRasterLayer.RasterCount
            if not bands==1:
                QMessageBox.information(None, "Erro no MDT", "Numero de bandas do Modelo Digital de Terreno seleccionado > 1. Nenhum calculo efectuado." )
            else:
                progress.setPercentage(0)
                progress.setText("A determinar pontos de seccionamento...")
                # Determinar tamanho do pixel do MDT
                geoTransf =theRasterLayer.GetGeoTransform()
                pixelSizeX = abs(geoTransf[1])
                pixelSizeY = abs(geoTransf[5])
                pixelSize = ((pixelSizeX + pixelSizeY) / 2)
                Comprimento_segmento = (Multiplicador_Comprimento_segmento * pixelSize)
                # Executar seccionamento da layer de linhas (caminhos)
                points = []
                progress.setPercentage(25)
                time.sleep(2)
                for feat in theLayerActive.getFeatures():
                    geom = feat.geometry()
                    length = geom.length()
                    iter = Comprimento_segmento
                    while iter <= (float(length)-float(Comprimento_segmento)):
                        pt = geom.interpolate(iter).exportToWkt()
                        points.append(pt)
                        iter += Comprimento_segmento
                epsg = theLayerActive.crs().authid()
                uri = "Point?crs=" + epsg + "&field=id:integer""&index=yes"
                mem_layer = QgsVectorLayer(uri,'SplitPoints', 'memory')
                prov = mem_layer.dataProvider()
                feats = [ QgsFeature() for i in range(len(points)) ]
                for i, feat in enumerate(feats):
                    feat.setAttributes([i])
                    feat.setGeometry(QgsGeometry.fromWkt(points[i]))
                prov.addFeatures(feats)
                QgsMapLayerRegistry.instance().addMapLayer(mem_layer)
                progress.setPercentage(50)
                progress.setText("A construir layer de linhas 'Intersection'...")
                time.sleep(2)
                Intersection=processing.runandload('saga:splitlinesatpoints', theLayerActive,mem_layer,1,0.001,None)
                # Verificar se o algoritmo  anterior  foi executado e se foi criada uma layer denominada  Intersection
                if len(QgsMapLayerRegistry.instance().mapLayersByName('Intersection')) == 0:
                    QMessageBox.information(None, "Erro!", "Nao foi possivel criar a layer de declives 'Intersection'. A terminar script." )
                    while len(QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')) > 0:
                        layer = QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')[0]
                        QgsMapLayerRegistry.instance().removeMapLayer(layer.id())
                else:
                    theactivelayer =  iface.activeLayer()
                    # Eliminar campo Declv
                    theactivelayer.startEditing()
                    theFields = []
                    for field in theactivelayer.fields():
                        idx = theactivelayer.fieldNameIndex('Declv')
                        theFields.append(idx)
                        theactivelayer.deleteAttributes(theFields)
                    theactivelayer.updateFields()
                    theactivelayer.commitChanges()
                    # Criar novo campo Declv 
                    theactivelayer.startEditing()
                    theactivelayer.dataProvider().addAttributes([QgsField("Declv",  QVariant.Double, "double", 15, 3)])
                    theactivelayer.updateFields()
                    theactivelayer.commitChanges()
                    progress.setPercentage(100)
                    time.sleep(2)
                    #    Determinar a altitude no MDT escolhido para as coordenadas dos pontos de inicio e fim de cada segmento   
                    progress.setPercentage(0)
                    progress.setText("A calcular declives...")
                    count = theactivelayer.featureCount()
                    bandeira1 = [0]
                    i = 0
                    features = theactivelayer.getFeatures()
                    for feature in features:
                        i = i + 1
                        percent = (i/float(count)) * 100
                        progress.setPercentage(percent)
                        geom = feature.geometry()
                        feat_geom = geom.asPolyline()
                        # x_st e y_st: coordenadas do ponto de inicio do segmento
                        x_st=feat_geom[0][0]
                        y_st=feat_geom[0][1]
                        # x_end e y_end: coordenadas do ponto de fim do segmento
                        x_end=feat_geom[-1][0]
                        y_end=feat_geom[-1][1]
                        geometryV2 = feature.geometry().geometry()
                        # Obter os valores de altitude no raster, nas coordenadas de inicio e fim de cada segmento
                        # Altitude no ponto de inicio
                        rasterx = int((x_st - geoTransf[0]) / geoTransf[1])
                        rastery = int((y_st - geoTransf[3]) / geoTransf[5])
                        # Valor da altitude nas coordenadas no inicio do segmento
                        elevationStart=theRasterLayer.GetRasterBand(1).ReadAsArray(rasterx,rastery, 1, 1)   # Altitude no inicio do segmento
                        # Altitude no ponto de fim
                        rasterx = int((x_end - geoTransf[0]) / geoTransf[1])
                        rastery = int((y_end - geoTransf[3]) / geoTransf[5])
                        # Valor da altitude nas coordenadas de fim do segmento
                        elevationEnd=theRasterLayer.GetRasterBand(1).ReadAsArray(rasterx,rastery, 1, 1)   # Altitude no fim do segmento
                        # Determinar o comprimento de cada segmento para despistar segmentos menores que 3*tamanho do pixel
                        geom = feature.geometry()
                        Comprimento = geom.length()
                        # Valor da elevacao
                        if (elevationStart < -99) :   # Valores <-99 considerados NO DATA
                            bandeira1=[1]
                            continue
                        elif (elevationEnd < -99) :
                            bandeira1=[1]
                            continue
                        elif (Comprimento < (pixelSize * 3)) :
                            bandeira1=[1]
                            continue
                        elevation = int(abs((elevationStart - elevationEnd)*10000))   # Multiplicado por 10000 para evitar perder precisao devido ao uso de Int
                        # Calcular declive do segmento com base no comprimento do segmento e na diferenca dos valores de z do inicio e do fim do segmento
                        theactivelayer.startEditing()
                        idx=theactivelayer.fieldNameIndex('Declv')
                        e=QgsExpression( ''+str(elevation)+'/($length * 100 )')
                        e.prepare(theactivelayer.pendingFields())
                        feature[idx]=e.evaluate(feature)
                        theactivelayer.updateFeature(feature)
                        theactivelayer.commitChanges()
                    if bandeira1[0] == 1:
                        QMessageBox.information(None, "Valores de declive nulos", "Dois casos se podem dar: 1-alguns (ou todos) pontos de inicio ou fim de segmento ficam em zonas do MDT com valores NO DATA, ou, 2-alguns elementos da layer de caminhos possuem um comprimento inferior a 3 vezes o tamanho da grid do MDT. Nestes casos deixaram-se em NULL os valores de declive.")
                    else:
                        QMessageBox.information(None, "Calcular declives v4d", "Script executado!")
#
while len(QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')) > 0:
    layer = QgsMapLayerRegistry.instance().mapLayersByName('SplitPoints')[0]
    QgsMapLayerRegistry.instance().removeMapLayer(layer.id())
#  Sobral Almeida, Setembro/2018

